############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import types, copy
from ZSI import TC, TCtimes
from ZSI.TC import _is_xsd_or_soap_ns
from ZSI import _copyright, _children, _child_elements, \
        _inttypes, _stringtypes, _seqtypes, _find_arraytype, _find_href, \
        _find_type, \
        EvaluateException 
from ZSI.TCcompound import _check_typecode_list,_find_arrayoffset , \
     _find_arrayposition,_offset_pat,_position_pat

from xml.dom.ext import SplitQName

DEBUG  = 0
def print_debug(msg, level=1, *l, **kw):
    if DEBUG >= level:
        for i in l:
            msg += '\n*\t' + str(i)
        for k,v in kw.items():
            msg += '\n**\t%s: %s' %(k,v)
        print 'TYPECODES(%d): %s' %(level,msg)

############################################################################
# Module Classes: BaseTypeInterpreter
###########################################################################

TC.Boolean.tag = 'boolean'
class BaseTypeInterpreter:
    """
    Example mapping of wsdl types to zsi python types.
    Checks against all available classes in ZSI.TC
    Used in TypeCodeSequenceGenerator, and TC_Template.
    """
    def __init__(self):
        self.__type_list = [TC.Boolean, TC.Iinteger, TC.IunsignedShort, TC.gYearMonth, TC.Base64String, \
                            TC.InonNegativeInteger, TC.Iint, TC.String, TC.gDateTime, \
                            TC.IunsignedInt, TC.IpositiveInteger, TC.FPfloat, TC.gDay, \
                            TC.InegativeInteger, TC.gDate, TC.URI, TC.HexBinaryString, \
                            TC.IunsignedByte, TC.gMonthDay, TC.InonPositiveInteger, \
                            TC.Ibyte, TC.FPdouble, TC.gTime, TC.gYear, TC.Ilong, \
                            TC.IunsignedLong, TC.Ishort]
        return
    
    def __map_type_to_typecode(self, msg_type):
        prefix, name = SplitQName(msg_type)

        for tc in self.__type_list:
             if tc.tag == name:
                 break
        else:
            tc = TC.Any
        return tc

    def get_typeclass(self, msg_type, targetNamespace):
        if _is_xsd_or_soap_ns(targetNamespace):
            return self.__map_type_to_typecode(msg_type)
        return None

    def get_typecode(self, name, msg_type, ns_prefix=''):
        typeC = self.__map_type_to_typecode(msg_type)
        tcs = []
        tcd = {}
        
        if typeC:
            tcs += [typeC(ns_prefix + name)]
            tcd.update({name:typeC})
        return  tcs,tcd

    def isBasicType(self, myType):
        prefix, name = SplitQName(myType)
        return (prefix == 'xsi') or (prefix == 'xsd')
            
    
class TC_Struct(TC.TypeCode):
    def __init__(self, pyclass, ofwhat, pname=None, inorder=0, inline=0,
    mutable=1, hasextras=0, nsdict={}, **kw):
        '''pyclass -- the Python class to hold the fields
        ofwhat -- a list of fields to be in the struct
        hasextras -- ignore extra input fields
        inorder -- fields must be in exact order or not
        inline -- dont href/id when serializing
        mutable -- object could change between multiple serializations
        type -- the (URI,localname) of the datatype
        '''
        TC.TypeCode.__init__(self, pname, **kw)
        self.pyclass = pyclass
        self.inorder = inorder
        self.inline = inline
        self.mutable = mutable
        if self.mutable: self.inline = 1
        self.hasextras = hasextras
        self.type = kw.get('type')
        t = type(ofwhat)

        if t not in _seqtypes:
            raise TypeError(
                'Struct ofwhat must be list or sequence, not ' + str(t))

        self.ofwhat, self.lenofwhat = tuple(ofwhat), len(ofwhat)
        if kw.has_key('typed'):
            t = kw['typed']
            for w in self.ofwhat: w.typed = t
        if TC.TypeCode.typechecks:
            if (self.pyclass != None) and\
	       (type(self.pyclass) != types.ClassType) and not\
	       (isinstance(self.pyclass, type)):
                raise TypeError('pyclass must be None or a class or a type(dict), not ' +
                        str(type(self.pyclass)))
            _check_typecode_list(self.ofwhat, 'Struct')
        self.nsdict = nsdict
    def parse(self, elt, ps):
        self.checkname(elt, ps)
        if self.type and \
        self.checktype(elt, ps) not in [ self.type, (None,None) ]:
            raise EvaluateException('Struct has wrong type', ps.Backtrace(elt))
        href = _find_href(elt)
        if href:
            if _children(elt):
                raise EvaluateException('Struct has content and HREF',
                        ps.Backtrace(elt))
            elt = ps.FindLocalHREF(href, elt)
        c = _child_elements(elt)
        count = len(c)
        if self.nilled(elt, ps): return None
        if count > self.lenofwhat and not self.hasextras:
            raise EvaluateException('Too many sub-elements', ps.Backtrace(elt))

        # Create the object.
        v = {}

        # Clone list of kids (we null it out as we process)
        c, crange = c[:], range(len(c))
        # Loop over all items we're expecting
        for i,what in [ (i, self.ofwhat[i]) for i in range(self.lenofwhat) ]:
            # Loop over all available kids
            for j,c_elt in [ (j, c[j]) for j in crange if c[j] ]:
                if what.name_match(c_elt):
                    # Parse value, and mark this one done.
                    try:
                        value = what.parse(c_elt, ps)
                    except EvaluateException, e:
                        e.str = '%s.%s: %s' % \
                                (self.pname or '?', what.aname or '?', e.str)
                        raise e
                    if what.repeatable and v.has_key(what.aname):
                        v[what.aname] = v[what.aname].append(value)
                    else:
                        v[what.aname] = [value]

                    c[j] = None
                    break
                # No match; if it was supposed to be here, that's an error.
                if self.inorder and i == j:
                    raise EvaluateException('Out of order struct',
                            ps.Backtrace(c_elt))
            else:
                if not what.optional:
                    raise EvaluateException('Element "' + what.aname + \
                        '" missing from struct', ps.Backtrace(elt))
                if hasattr(what, 'default'):
                    v[what.aname] = what.default

        if not self.hasextras:
            extras = [c_elt for c_elt in c if c_elt]
            if len(extras):
                raise EvaluateException('Extra sub-elements (' + \
                        ','.join([c.nodeName for c in extras]) + ') found',
                        ps.Backtrace(elt))

        if not self.pyclass: return v

        try:
            pyobj = self.pyclass(self.aname)
        except Exception, e:
            raise TypeError("Constructing %s: %s" % (self.pname, str(e)))
        for key in v.keys():
            setattr(pyobj, key, v[key])
        return pyobj
    def serialize(self, sw, pyobj, inline=None, name=None, attrtext='', **kw):
        if inline or self.inline:
            self.cb(sw, pyobj, **kw)
        else:
            objid = '%x' % id(pyobj)
            n = name or self.oname or ('E' + objid)
            print >>sw, '<%s%s href="#%s"/>' % (n, attrtext, objid)
            sw.AddCallback(self.cb, pyobj)
    def cb(self, sw, pyobj, name=None, **kw):
        if not self.mutable and sw.Known(pyobj): return
        objid = '%x' % id(pyobj)
        n = name or self.oname or ('E' + objid)

        if self.inline:
            print >>sw, '<%s' % n     
        else:
            print >>sw, '<%s id="%s"' % (n, objid)

        """

        #HACK NSDICT IN  sw.writeNSdict(nsdict)
        try:
            pyobj.nsdict
            print >>sw, ' %s>' % pyobj.nsdict    
        except AttributeError, ex:
            print_debug('AttributeError ', 5, *ex.args)
            print >>sw, '>'
        """
        print >>sw, '>'

        
        if self.pyclass:
            d = pyobj.__dict__
        else:
            d = pyobj
            if TC.TypeCode.typechecks and type(d) != types.DictType:
                raise TypeError("Classless struct didn't get dictionary")
        for what in self.ofwhat:
            v = d.get(what.aname)
            if what.optional and v == None: continue
            try:
                if what.repeatable and type(v) in _seqtypes:
                    for v2 in v:
                        what.serialize(sw, v2)
                else:
                    what.serialize(sw, v)
            except Exception, e:
                #raise RuntimeError('Serializing %s.%s, %s %s' %
                #        (n, what.aname or '?', e.__class__.__name__, str(e)))
                raise
        print >>sw, '</%s>' % n



class ComplexType(TC_Struct):
    def __init__(self, pyclass, ofwhat, pname=None, inorder=0, inline=0,
    mutable=1, hasextras=0, nsdict={}, **kw):
        '''pyclass -- the Python class to hold the fields
        ofwhat -- a list of fields to be in the struct
        hasextras -- ignore extra input fields
        inorder -- fields must be in exact order or not
        inline -- dont href/id when serializing
        mutable -- object could change between multiple serializations
        type -- the (URI,localname) of the datatype
        '''
        TC.TypeCode.__init__(self, pname, **kw)
        self.pyclass = pyclass
        self.inorder = inorder
        self.inline = inline
        self.mutable = mutable
        if self.mutable: self.inline = 1
        self.hasextras = hasextras
        self.type = kw.get('type')
        t = type(ofwhat)

        if t not in _seqtypes:
            raise TypeError(
                'Struct ofwhat must be list or sequence, not ' + str(t))

        self.ofwhat, self.lenofwhat = tuple(ofwhat), len(ofwhat)
	""" If an item is repeatable, lenofwhat is irrelevant
        """
        for ct in ofwhat:
            if hasattr(ct, 'repeatable') and ct.repeatable:
                self.hasextras = 1
                break

        if kw.has_key('typed'):
            t = kw['typed']
            for w in self.ofwhat: w.typed = t
        if TC.TypeCode.typechecks:
            if (self.pyclass != None) and\
	       (type(self.pyclass) != types.ClassType) and not\
	       (isinstance(self.pyclass, type)):
                raise TypeError('pyclass must be None or a class or a type(dict), not ' +
                        str(type(self.pyclass)))
            _check_typecode_list(self.ofwhat, 'ComplexType')
        self.nsdict = nsdict


    def serialize(self, sw, pyobj, inline=None, name=None, attrtext='', **kw):
        if not self.mutable and sw.Known(pyobj): return
        objid = '%x' % id(pyobj)
        n = name or self.oname or ('E' + objid)
        if self.inline:
            print >>sw, '<%s>' % n
        else:
            print >>sw, '<%s id="%s">' % (n, objid)

        content = copy.deepcopy(pyobj.content)
        for i in range(len(content)):
            for k,v in content[i].__dict__.items(): print "\t%s = %s" %(k,v)
	    if inline or self.inline:
                self.ofwhat[i].serialize(sw, content[i])
	    else:
		#objid = '%x' % id(pyobj.content[i])
		#n = name or self.oname or ('E' + objid)
		#print >>sw, '<%s%s href="#%s"/>' % (n, attrtext, objid)
		#sw.AddCallback(self.cb, i)
                pass

        # ignore the xmlns if it was explicitly stated
        i = n.find('xmlns')
        if i > 0:
            print >>sw, '</%s>' % n[:i - 1]
        else:
            print >>sw, '</%s>' % n

    def parse(self, elt, ps):
        self.checkname(elt, ps)
        if self.type and \
        self.checktype(elt, ps) not in [ self.type, (None,None) ]:
            raise EvaluateException('Struct has wrong type', ps.Backtrace(elt))
        href = _find_href(elt)
        if href:
            if _children(elt):
                raise EvaluateException('Struct has content and HREF',
                        ps.Backtrace(elt))
            elt = ps.FindLocalHREF(href, elt)

        c = _child_elements(elt)
        count = len(c)
        if self.nilled(elt, ps): return None
        if count > self.lenofwhat and not self.hasextras:
            raise EvaluateException('Too many sub-elements', ps.Backtrace(elt))

        # Create the object.
        v = {}

        # Clone list of kids (we null it out as we process)
        c, crange = c[:], range(len(c))
        #clone = elt.cloneNode(1)

        # Loop over all items we're expecting
        for i,what in [ (i, self.ofwhat[i]) for i in range(self.lenofwhat) ]:
            print "OFWHAT: ", what.ofwhat
            v2 = what.parse(c, ps)
            #v2 = what.parse(elt, ps)
            print "V2: ", v2
            for key,value in v2.items():
                if v.has_key(key):
                    v[key] += v2[key]
                else:
                    v[key] = v2[key]
        """
        if not self.hasextras:
            extras = [c_elt for c_elt in c if c_elt]
            if len(extras):
                raise EvaluateException('Extra sub-elements (' + \
                        ','.join([c.nodeName for c in extras]) + ') found',
                        ps.Backtrace(elt))
        """

        if not self.pyclass: return v

        """  Here contentTypes do not have names.
        """
        try:
            #pyobj = self.pyclass(self.aname)
            pyobj = self.pyclass()
        except Exception, e:
            raise TypeError("Constructing %s: %s" % (self.pname, str(e)))
        for key in v.keys():
            setattr(pyobj, key, v[key])
        return pyobj


class Choice(TC.TypeCode):
    '''A union, also known as an XSD choice; pick one.  repeatable, optional only
       valid args in content type choice only.  
       local <element> declaratoins in a choice can not be specified as repeatable/optional.
    '''

    def __init__(self, pyclass, ofwhat, pname=None, maxOccurs=1, minOccurs=1, **kw):
        '''choices -- list of typecodes; exactly one must be match
           ofwhat -- a list of fields to be in the struct
        '''
        self.pyclass = pyclass
        self.maxOccurs = maxOccurs
        TC.TypeCode.__init__(self, pname, repeatable=maxOccurs>1, optional=not minOccurs, **kw)
        t = type(ofwhat)
        if t not in _seqtypes:
            raise TypeError(
                'Struct ofwhat must be list or sequence, not ' + str(t))

        self.ofwhat = {}
        self.lenofwhat = 0
        for what in ofwhat:
            self.ofwhat[what.aname] = what
            self.lenofwhat += 1

	""" If an item is repeatable, lenofwhat is irrelevant
        """
        for ct in ofwhat:
            #if hasattr(ct, 'maxOccurs'):
            #    self.lenofwhat += ct.maxOccurs
            if hasattr(ct, 'repeatable') and ct.repeatable:
                self.hasextras = 1
                break


        if len(self.ofwhat) == 0: raise TypeError('Empty choice')
        _check_typecode_list(self.ofwhat.values(), 'Choice')

    def parse(self, c, ps):
        """The parent <ComplexType> needs to call choice.parse w/ its
           root node.
           pyobj -- set variable to this passed in object
        """
        count = len(c)

        # Create the object.
        v = {}

        """ null out kids, and once we reach
            an element we don't understand return.
        """
        ofwhat = self.ofwhat.values()
        found = 0

	while count > 0:
	    if self.ofwhat.has_key(c[0].localName) and found < self.maxOccurs:
		what = self.ofwhat[c[0].localName]

		try:
		    value = what.parse(c[0], ps)
		except EvaluateException, e:
		    e.str = '%s.%s: %s' % \
			(self.pname or '?', what.aname or '?', e.str)
		    raise e

		if self.repeatable and v.has_key(what.aname):
		    v[what.aname].append(value)
		else:
                    v[what.aname] = [value]
		    
		del c[0]
		found += 1
                count -= 1
	    else:
		return v
	else:
	    if not self.optional and not found:
		raise EvaluateException('Choice is required') 

        return v

    def serialize(self, sw, pyobj, **kw):
        """ If specify repeatable and using sequential type for 
            a pyobj data member you must store it in a 
            tuple or list.
        """
        if self.pyclass:
            d = pyobj.__dict__
        else:
            d = pyobj
            if TC.TypeCode.typechecks and type(d) != types.DictType:
                raise TypeError("Classless struct didn't get dictionary")
        
        if not len(d) and not self.optional: 
            raise RuntimeError, 'no data for required choice'

        elif self.repeatable: 
            occurs = 0
            for k,v in d.items():
                if not self.ofwhat.has_key(k): continue
                what = self.ofwhat[k] 
                if type(v) in _seqtypes:
                    for v2 in v:
                        what.serialize(sw, v2)
                        occurs += 1
                else:
                    what.serialize(sw, v)
                    occurs += 1
                #delattr(pyobj, k)
            else:
                if self.optional > occurs or occurs > self.maxOccurs:
                    raise RuntimeError, 'Invalid number of choice occurances (%d)' %occurs

        elif len(d) == 1:
            k,v = d.items()[0]
            what = self.ofwhat[k] 
            what.serialize(sw, v)
            #delattr(pyobj, k)

        else:
            raise RuntimeError, 'problem serializing'

    def name_match(self, elt):
        for o in self.ofwhat.values():
            if o.name_match(elt): return 1
        return 0


class All(TC.TypeCode):
    """ <all maxOccurs=1, minOccurs=0,1>
           <element maxOccurs=1, minOccurs=0,1/>
        </all>
    """

    def __init__(self, pyclass, ofwhat, pname=None, minOccurs=1, **kw): pass





class Sequence(TC.TypeCode):
    """ <sequence maxOccurs=unbounded, minOccurs=0,1>
           <element maxOccurs=1, minOccurs=0,1/>
        </sequence>
    """
    def __init__(self, pyclass, ofwhat, pname=None, maxOccurs=1, minOccurs=1, **kw):
        """choices -- list of typecodes; exactly one must be match
           ofwhat -- a list of fields to be in the struct
        """
        self.pyclass = pyclass
        self.maxOccurs = maxOccurs
        TC.TypeCode.__init__(self, pname, repeatable=maxOccurs>1, optional=not minOccurs, **kw)
        t = type(ofwhat)
        if t not in _seqtypes:
            raise TypeError(
                'Struct ofwhat must be list or sequence, not ' + str(t))

        """ Need to maintain order for sequence
        """
        self.ofwhat = ofwhat
        self.lenofwhat = len(ofwhat)

	""" <sequence>  minOccurs maxOccurs
            <element>  repeatable
        """
        for elm in ofwhat:
            if hasattr(elm, 'repeatable') and elm.repeatable:
                self.hasextras = 1
                break

        if len(self.ofwhat) == 0: raise TypeError('Empty choice')
        _check_typecode_list(self.ofwhat, 'Choice')

    def parse(self, c, ps):
        """
        c = _child_elements(elt)
        count = len(c)
        if self.nilled(elt, ps): return None
        #if count > self.lenofwhat and not self.hasextras:
        #    raise EvaluateException('Too many sub-elements', ps.Backtrace(elt))
        """
        count = len(c)

        # Create the object.
        v = {}


        # Loop over all items we're expecting
        ofwhat = self.ofwhat
        count = len(c)
        crange = range(count)

        irange = range(self.lenofwhat)

        occurs = 0
        for o in range(self.maxOccurs):
            found = 0
            notfound = []
	    for i,what in [ (i, ofwhat[i]) for i in range(self.lenofwhat) ]:
	        for j,element in [ (j, c[j]) for j in range(len(c)) ]:
		    if c[j] and what.name_match(c[j]):
			value = what.parse(c[j], ps)
		    else:
			continue

		    if self.repeatable and v.has_key(what.aname):
		        v[what.aname].append(value)
		    else:
		        v[what.aname] = [value]
		    
		    del c[j]
                    found = 1
		    break
		else:
		    if not what.optional:
                        notfound.append((i,what))
                        if found:
                            # catchable exception in parent
			    raise EvaluateException('Element "' + what.aname + \
                                      '" missing from sequence', ps.Backtrace(c.parentNode))
            else:
                # catchable exception in parent
                if notfound and found:
		    raise EvaluateException('Sequence out of order, not found %s' %(notfound))
                elif notfound:
                    # all done
                    break
                occurs += found
                
        else:
            if occurs < self.optional:
		raise EvaluateException('sequence missing')

        return v

    def serialize(self, sw, pyobj, **kw):
        """ If specify repeatable and using sequential type for 
            a pyobj data member you must store it in a 
            tuple or list.
        """
        if self.pyclass:
            d = pyobj.__dict__
        else:
            d = pyobj
            if TC.TypeCode.typechecks and type(d) != types.DictType:
                raise TypeError("Classless struct didn't get dictionary")
        
        #if not len(d) and not self.optional: 
        #    raise RuntimeError, 'no data for required sequence'

        occurs = 0
        for i in range(self.maxOccurs):
            for what in self.ofwhat:
                if not d.has_key(what.aname) and (what.optional or i):
                    continue

                v = d[what.aname] 
                if type(v) in _seqtypes:
                    what.serialize(sw, v[0])
                    del v[0]
                    if len(v) == 0:
                        del d[what.aname]
                else:
                    what.serialize(sw, v)
                    del d[what.aname]
                occurs += 1

                #delattr(pyobj, k)
        else:
            if self.optional > occurs or occurs > self.maxOccurs:
                raise RuntimeError, 'Invalid number of choice occurances (%d)' %occurs



    def name_match(self, elt):
        for o in self.ofwhat.values():
            if o.name_match(elt): return 1
        return 0



        
###########################################################################
# Module Classes: Type Codes
#                 ZSI.TC
#                 pyGMA
#
##########################################################################
class TC_Base(TC.TypeCode):
    name2class = {}
    def parse(self, elt, ps):
        return self.choice.parse(elt, ps)[1]
    
    def __getattr__(self, attr):
        if attr == 'choice':
            l = []
            for k,v in self.__class__.name2class.items():
                print_debug('TC_Base: (%s)\n%s' %(v.typecode,v.typecode.__dict__))
                l += [v.typecode]
            choice = TC.Choice(l, 'Item')
            self.__dict__['choice'] = choice
            return choice
        raise AttributeError(attr)
    
class TC_Template(TC.TypeCode):
    """
    nspname
    pname
    aname
    oname
    repeatable
    typed
    unique
    optional
    """
    def __init__(self, name, **seq):
        TC.TypeCode.__init__(self)
        self.pname = name
        typeInt = BaseTypeInterpreter()
        for n,v in seq.items():
            self.tc = typeInt.get_typecode(n, v)
            
    def parse(self, elt, ps):
        self.checkname(elt, ps)
        if self.nilled(elt, ps): return None
        p = self.tc.parse
        v = {}
        d = None
        for c in _child_elements(elt):
            d = p(c, ps)
        return d

class TC_NOTYPE(TC.String):
    def serialize(self, sw, pyobj, name=None, attrtext='', **kw):
        objid = '%x' % id(pyobj)
        n = name or self.oname or ('E' + objid)
        print >>sw, '<%s>%s</%s>' % (n, pyobj, n)
        return
