#!/bin/env python
############################################################################
# Keith Beattie <KSBeattie@lbl.gov>, LBNL
# See Copyright for copyright notice!
###########################################################################
import getopt, re, sys, socket, time, threading

"""

GridSniffer:

A tcpmon-style 'sniffer' logging the upstream and downstream messages
exchanged between a tcp client and server.

If -r flag is set URLs of the form 'http://<serverIP>:<serverPort>/'
(where <serverIP> is the IP of the server) are re-written to
'http://<localIP>:<listenPort>/', In this way we stay in the loop on
messages exachanged.


Arch:

This works by creating two threads, one for listening to upstream
(client to server) data and another for listening to downstream data
(server to client).

GridSniffer loops forever accepting connections from client.  Upon accept:
  - Connects to server
  - Spawns upstream and downstream threads
  - Waits for them to finish

UpStream thread:
  - reads data from client
  - prints it to log
  - sends it to server
  - when connection is closed, shutdown downstream (client) socket.

DownStream thread:
  - reads data from server
  - possibly re-writes URLs
  - prints it to log
  - sends it to client
  - when connection is close, shutdown upstream (server) socket.


Wish List:

  - Expand the regexp for re-writes to replace more than just the
  'http://serverIP:serverPort/' case (though that's the only one I've
  seen from GRID containers).  Though HTTP isn't a priority here, it
  doesn't handle '301 Moved Permanently' warnings.

  - If the kinds of re-writes that are done is expanded (as suggested
  above) consider breaking the re-writing out to a
  DownStream.rewrite() method as the repl arg of the re.subn() call.
  This way details of what is being re-written can be printed to
  stdout.

  - Give the log files better names?

  - Is it possible that a redirect is not an indication that the
  current connection will be closing very soon?  I assume that if
  there is a re-direct, then the current connection will be shutdown
  and a new one will be initiated by the client.  Perhaps it would be
  better to spawn a whole new GridSniffer thread for each re-write or
  each connection?  Ugh, that'd be a big change and it wold mess up
  the logging.  This is beyond what I think this tool is going to be
  used for.

  - Speaking of big changes, it might be better to switch to the
  'select model' (as opposed to the current 'thread model') and allow
  multiple accepts to be open at once.  Then we'd wait on messages from any
  of the sockets and track (exactly how I'm not sure yet) which one
  sends to which.  This would be a better strategy if we need to
  maintain multiple connections to servers in order to stay
  'in-the-middle'.


"""

class UpStream(threading.Thread):

    """ The up-stream thread reading from the client printing to the
    log and sending to the server. """

    def __init__(self, csock, ssock, log):
        threading.Thread.__init__(self)
        self.client = csock
        self.server = ssock
        self.log    = log

        # Log timestamp, etc. about new connection
        now = time.time()
        peer = csock.getpeername()
        print >>log, "===== GridSniffer: Connection from %s:%s at %s:%08.5f =====" % (
            peer[0], peer[1], time.strftime('%Y %a %b %d %H:%M', time.localtime(now)), now % 60)

    def run(self):

        """ Read from client, log it and send it to server.  Return
        when the client socket is closed, but first shutdown the
        server socket as a notification to the DownStream thread to
        finish. """

        try:
            while 1:
                data = self.client.recv(8192)
                print "gs.UpStream: read %s bytes" % (len(data))
                if not data:
                    self.server.shutdown(2)  # This makes the DownStream thread stop
                    break
                print >>self.log, data,
                self.log.flush()
                self.server.sendall(data)
        except socket.error, e:
            print "gs.UpStream.run(): socket.error:", e


class DownStream(threading.Thread):

    """ The down-stream thread reading from the server printing, doing
    re-writes, printing to the log and sending to the client. """

    def __init__(self, csock, ssock, log, rewriteRE, localIP, listenPort):
        threading.Thread.__init__(self)
        self.client     = csock
        self.server     = ssock
        self.log        = log
        self.rewriteRE  = rewriteRE
        self.localIP    = localIP
        self.listenPort = listenPort

        # Log timestamp, etc. about new connection
        now = time.time()
        peer = ssock.getpeername()
        print >>log, "===== GridSniffer: Connection to %s:%s at %s:%08.5f =====" % (
            peer[0], peer[1], time.strftime('%Y %a %b %d %H:%M', time.localtime(now)), now % 60)

    def run(self):

        """ Read from server, log it and send it to client.  Return
        when the server socket is closed, but first shutdown the
        client socket as a notification to the UpStream thread to
        finish. """


        try:
            while 1:
                data = self.server.recv(8192)
                print "gs.DownStream: read %s bytes" % (len(data))
                if not data:
                    self.client.shutdown(2)  # To stop the UpStream thread
                    break
                if self.rewriteRE:
                    (data, count) = self.rewriteRE.subn("http://%s:%d/" % (self.localIP,
                                                                           self.listenPort),
                                                        data)
                    print "gs.DownStream: %d re-writes" % (count)
                print >>self.log, data,
                self.log.flush()
                self.client.sendall(data)
        except socket.error, e:
            print "gs.DownStream: socket.error:", e


class GridSniffer:

    """ Management class for the upstream and downstream threads.
    Opens and closes, log files and sockets. """

    def __init__(self, localIP, listenPort, server, serverPort, rewriteRE):
        self.localIP    = localIP
        self.listenPort = listenPort
        self.server     = server
        self.serverPort = serverPort
        self.rewriteRE  = rewriteRE

        self.gssock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.gssock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.gssock.bind(('', listenPort))
        self.gssock.listen(5)

    def go(self):

        """ Main GridSniffer event loop.
        
        Loop accepting connections from clients.  Upon accect
          - connect to server
          - spawn threads
          - wait for them to finish

        """

        # open log files
        uplog = open("gs_up.log", "w")
        downlog = open("gs_down.log", "w")

        try:
            while 1:
                # Wait for incomming connection:
                csock,caddr = self.gssock.accept()
                now = time.time()
                print "gs: Connection from %s:%s %s:%08.5f" % (caddr[0], caddr[1],
                                                               time.strftime('%Y %a %b %d %H:%M',
                                                                             time.localtime(now)),
                                                               now % 60)
                
                # Connect to server
                ssock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                ssock.connect((self.server, self.serverPort))

                # Start upstream thread
                upstream = UpStream(csock, ssock, uplog)
                upstream.start()
                
                # Start downstream thread
                downstream = DownStream(csock, ssock, downlog, self.rewriteRE,
                                        self.localIP, self.listenPort)
                downstream.start()

                # Wait for threads to finish (and they both will
                # because if one does it shuts down the other's socket
                # and eventually one of their sockets will be shutdown
                # via the other end.)
                downstream.join()
                upstream.join()
                
                # Close sockets
                csock.close()
                ssock.close()

                print "gs: Connection closed, looping to accept new client."

                
        except KeyboardInterrupt, e:
            pass

        # Close log files
        uplog.close()
        downlog.close()

#### End of class definitions and begining of main program ####

# Usage string
usage = """Usage: gridSniffer -l listenPort -s server:port [-r] [-h]
  where:
    listenPort  -> The port gridSniffer is listening on (the one your client connects to).
    server:port -> The server and port running the grid container.
    -r          -> re-write URLs from server:port to localhost:listenPort.
    -h          -> prints this message and exits.
$Id: gridSniffer.py,v 1.4 2003/04/24 19:27:30 ksb Exp $"""

if __name__ == "__main__":
    # Check command line args
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hrl:s:")
    except getopt.GetoptError:
        print usage
        sys.exit(2)

    listenPort = server = serverPort = None
    rewrite    = 0
    try:
        for o, a in opts:
            if o == '-h':
                print usage
                sys.exit()
            elif o == '-r':
                rewrite = 1
            elif o == '-l':
                listenPort = int(a)
            elif o == '-s':
                server, serverPort = a.split(':')
                serverPort = int(serverPort)
    except ValueError, e:
        print usage
        sys.exit(2)

    if listenPort == None or server == None or serverPort == None:
        print usage
        sys.exit(2)

    # Calc the regular expression for URL re-writing
    localIP = socket.gethostbyname(socket.gethostname())
    serverIP = socket.gethostbyname(server)
    if rewrite:
        rewriteRE = re.compile("http://%s:%d/" % (serverIP, serverPort))
    else:
        rewriteRE = None

    # Start the GridSniffer
    gs = GridSniffer(localIP, listenPort, server, serverPort, rewriteRE)
    print "GridSniffer: %d <-> %s:%d re-writes=%s" % (listenPort, server, serverPort, rewrite)
    gs.go()
